import re
import sys
import time
import pyext

from athenaCL.libATH.sieve import *

MODE_SIMPLIFY  = 0
MODE_DECOMPOSE = 1

def powerfactorise(x):
	""" factorise x into its prime powers """
	factors = []
	factor = 2
	while factor <= x:
		if x % factor == 0:
			if len(factors) > 0 and factors[-1] % factor == 0:
				factors[-1] *= factor
			else:
				factors.append(factor)
			x /= factor
		else:
			factor += 1
	return factors

class pysieve(pyext._class):
	_inlets = 1
	_outlets = 2
	
	def __init__(self):
		self.clear()
		
		self.mode = MODE_SIMPLIFY

		self.display_limit = 88
		
		self.period        = self.display_limit;
		self.period_auto   = self.period
		self.period_manual = self.period
		
		
		self.autoperiod = False
		
		self.points = []
		self.bits = self.pointstobits(self.points)
		
		print "sieve: __init__"
	
	def bang_(self, n):
		print "sieve: hello world"
 	
	def union_(self, inlet, *l):
		modulo = int(l[0])
		residue = int(l[1])
		print "%d, %d" % (modulo, residue)
		self.str = "%s|%d@%d" % (self.str, modulo, residue)
		print "str: %s" % self.str
		self.update()
	
	def clear_(self, *l):
		self.points = []
		self.bits = self.pointstobits(self.points)
		self.update_points()
		self.update_segments()
		
	def formula_(self, inlet, *l):
		text = l[0]
		try:
			self.str = str(text)

			# we should really generate the points for the entirety
			# of the sieve's period, otherwise the later CompressionSeg gives
			# weird results (as it uses the points generated below), but we
			# don't know the period ahead of time...
			sieve = Sieve(self.str, range(self.display_limit))
			
			self.period_auto = sieve.period()
			if self.period_auto < 0:
				self.period_auto = self.period_manual
				
			self.points = sieve()
			self.bits = self.pointstobits(self.points)
			
			self.update_period()
			self.update_points()
			self.update_segments()
			
			self._outlet(1, "error", 0)
			print "period %d" % self.period_auto
			self._outlet(1, "period", self.period_auto)

		except Exception, e:
			print "exception: %s" % e
			self._outlet(1, "error", 1)

	def bits_(self, inlet, *l):
		if self.bits != l:
			self.bits = l
			self.points = self.bitstopoints(self.bits)
			self.update_formula()
			self.update_points()
	
	def period_(self, inlet, *l):
		self.period_manual = int(l[0])

		# CRASH WARNING
		# seems we must update points before segments...
		# swapping update_points/update_segments round causes rainbow wheel of death.
		self.update_period()
		self.update_points()
		self.update_segments()
	
	def autoperiod_(self, inlet, auto):
		auto = True if auto else False
		self.autoperiod = auto

		self.update_period()
		print "using period %d" % self.period
		return
		self.update_points()
		self.update_segments()
	
	def mode_(self, inlet, *l):
		self.mode = int(l[0])
		self.update_formula()
	
	def copyback_(self, inlet, *l):
		formula = self.generate_formula()
		self._outlet(1, "copyback", formula)
	
	def substitute_(self, inlet, a, b):
		formula = self.generate_formula()
		formula = re.sub("\\b%s(?=@)" % str(a), str(b), formula)
		self._outlet(1, "formula", formula)
		self._outlet(1, "copyback", formula)
	
	def bitshift_(self, inlet, distance):
		distance = -distance
		self.bits = self.bits[distance:] + self.bits[:distance]
		self.points = self.bitstopoints(self.bits)
		self.update_formula()
		self.update_points()
		self.update_segments()

	def pointstobits(self, points):
		""" convert a list of points to a list of bits:
		    [ 0, 2, 3, 5 ] -> [ 1, 0, 1, 1, 0, 1 ]"""
		bits = [0] * self.display_limit
		print "bits: %s" % bits
		print "points: %s" % points
		for point in points:
			bits[point] = 1
		print "bits: %s" % bits
		return bits

	def bitstopoints(self, bits):
		""" convert a list of points to a list of bits:
		    [ 1, 0, 1, 1, 0, 1 ] -> [ 0, 2, 3, 5 ] """
		points = map(lambda n: n[0], filter(lambda n: n[1], [ (key, value) for key, value in enumerate(bits) ]))
		return points

	def clear(self):
		self.str = "9999@9998"

	def generate_formula(self):
		formula = "-"
		
		if self.mode == MODE_SIMPLIFY:
			if len(self.points) > 1:
				z = range(0, self.period)
				residuals = CompressionSeg(self.points, z)()
				# added range(self.period)
				sieves = map(lambda r: Sieve(r.__str__(), range(self.period)), residuals)
				reduced = reduce(lambda x, y: x | y, sieves)
				formula = str(reduced)
		else:
			factors = powerfactorise(self.period)
			print "factors: %s" % factors
			residuals = []
			for point in self.points:	
				disjunctives = map(lambda n: "{%d@%d}" % (n, point % n), factors)
				residual = "{%s}" % "&".join(disjunctives)
				residuals.append(residual)
				
			formula = "|".join(residuals)
				
		return formula
	
	def update_period(self):
		if self.autoperiod:
			self.period = self.period_auto
		else:
			self.period = self.period_manual
		
	def update_formula(self):
		formula = self.generate_formula()
		self._outlet(1, "formula", formula)
		
	def update_points(self):
		if not self.points:
			return

		# format as indexed points (index, value), for coll:
		# [ (1, 0), (2, 4), (3, 8), (4, 12) ... ]
		indexedpoints = [ (x[0] + 1, x[1]) for x in enumerate(self.points) ]
		# add final point to represent wrap
		# indexedpoints = indexedpoints + [ (len(self.points), self.period) ]
		indexedpoints = list(reduce(lambda a, b: a + b, indexedpoints))
		print "indexedpoints: %s" % indexedpoints
		self._outlet(1, "points", indexedpoints)

	def update_segments(self):
		print "updating period = %d" % self.period
		width = self.period
		if width > self.display_limit:
			width = self.display_limit
		self._outlet(1, "bitwidth", width)

		if not self.points:
			return

		# format as indexed bits (index, on/off), for multislider:
		# [ (1, 0), (2, 0), (3, 1), (4, 0) ... ]
		indexedbits = [ (x[0] + 1, x[1]) for x in enumerate(self.bits) ]
		indexedbits = list(reduce(lambda a, b: a + b, indexedbits))			
		self._outlet(2, "select", indexedbits)
	